# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re
import unittest

from yaml2ical import ical
from yaml2ical import meeting
from yaml2ical.tests import sample_data


class MeetingTestCase(unittest.TestCase):

    def test_bad_meeting_day(self):
        self.assertRaises(ValueError,
                          meeting.load_meetings,
                          sample_data.BAD_MEETING_DAY)

    def test_load_yaml_file(self):
        m = meeting.load_meetings(sample_data.WEEKLY_MEETING)[0]
        self.assertEqual('OpenStack Subteam Meeting', m.project)
        self.assertEqual('Joe Developer', m.chair)
        self.assertEqual('Weekly meeting for Subteam project.\n',
                         m.description)
        self.assertEqual('Joe Developer', m.schedules[0].chair)

    def test_chair_overrides(self):
        m = meeting.load_meetings(sample_data.ALTERNATING_CHAIRS_MEETING)[0]
        self.assertEqual('OpenStack Subteam Meeting', m.project)
        self.assertIsNone(m.chair)
        self.assertEqual('Weekly meeting for Subteam project.\n',
                         m.description)
        self.assertEqual('Jane Developer', m.schedules[0].chair)
        self.assertEqual('John Developer', m.schedules[1].chair)

    def should_be_conflicting(self, yaml1, yaml2):
        """Exception is raised when meetings should conflict."""
        meeting_one = meeting.load_meetings(yaml1)
        meeting_two = meeting.load_meetings(yaml2)
        meeting_list = [meeting_one.pop(), meeting_two.pop()]
        self.assertRaises(meeting.MeetingConflictError,
                          meeting.check_for_meeting_conflicts,
                          meeting_list)

    def should_not_conflict(self, yaml1, yaml2):
        """No exception raised when meetings shouldn't conflict."""
        meeting_one = meeting.load_meetings(yaml1)
        meeting_two = meeting.load_meetings(yaml2)
        meeting_list = [meeting_one.pop(), meeting_two.pop()]
        meeting.check_for_meeting_conflicts(meeting_list)

    def test_weekly_conflict(self):
        self.should_be_conflicting(
            sample_data.WEEKLY_MEETING,
            sample_data.CONFLICTING_WEEKLY_MEETING)
        self.should_not_conflict(
            sample_data.WEEKLY_MEETING,
            sample_data.WEEKLY_OTHER_CHANNEL_MEETING)

    def test_biweekly_conflict(self):
        self.should_be_conflicting(
            sample_data.WEEKLY_MEETING,
            sample_data.ALTERNATING_MEETING)
        self.should_not_conflict(
            sample_data.ALTERNATING_MEETING,
            sample_data.BIWEEKLY_EVEN_MEETING)
        self.should_be_conflicting(
            sample_data.ALTERNATING_MEETING,
            sample_data.BIWEEKLY_ODD_MEETING)
        self.should_not_conflict(
            sample_data.BIWEEKLY_ODD_MEETING,
            sample_data.BIWEEKLY_EVEN_MEETING)

    def test_late_early_conflicts(self):
        self.should_be_conflicting(
            sample_data.MEETING_SUNDAY_LATE,
            sample_data.MEETING_MONDAY_EARLY)
        self.should_be_conflicting(
            sample_data.MEETING_MONDAY_LATE,
            sample_data.MEETING_TUESDAY_EARLY)

    def test_quadweekly_conflicts(self):
        self.should_be_conflicting(
            sample_data.BIWEEKLY_EVEN_MEETING,
            sample_data.QUADWEEKLY_MEETING_ALTERNATING)
        self.should_be_conflicting(
            sample_data.BIWEEKLY_EVEN_MEETING,
            sample_data.QUADWEEKLY_MEETING)
        self.should_be_conflicting(
            sample_data.BIWEEKLY_EVEN_MEETING,
            sample_data.QUADWEEKLY_MEETING_ALTERNATE)
        self.should_be_conflicting(
            sample_data.BIWEEKLY_EVEN_MEETING,
            sample_data.QUADWEEKLY_MEETING_WEEK_2)
        self.should_not_conflict(
            sample_data.BIWEEKLY_EVEN_MEETING,
            sample_data.QUADWEEKLY_MEETING_WEEK_1)
        self.should_not_conflict(
            sample_data.BIWEEKLY_EVEN_MEETING,
            sample_data.QUADWEEKLY_MEETING_WEEK_3)
        self.should_be_conflicting(
            sample_data.BIWEEKLY_EVEN_MEETING,
            sample_data.QUADWEEKLY_MEETING)
        self.should_be_conflicting(
            sample_data.BIWEEKLY_ODD_MEETING,
            sample_data.QUADWEEKLY_MEETING_WEEK_1)
        self.should_be_conflicting(
            sample_data.BIWEEKLY_ODD_MEETING,
            sample_data.QUADWEEKLY_MEETING_WEEK_3)
        self.should_not_conflict(
            sample_data.BIWEEKLY_ODD_MEETING,
            sample_data.QUADWEEKLY_MEETING)
        self.should_not_conflict(
            sample_data.BIWEEKLY_ODD_MEETING,
            sample_data.QUADWEEKLY_MEETING_WEEK_2)
        self.should_not_conflict(
            sample_data.BIWEEKLY_ODD_MEETING,
            sample_data.QUADWEEKLY_MEETING_ALTERNATING)
        self.should_not_conflict(
            sample_data.QUADWEEKLY_MEETING,
            sample_data.QUADWEEKLY_MEETING_WEEK_1)
        self.should_not_conflict(
            sample_data.QUADWEEKLY_MEETING,
            sample_data.QUADWEEKLY_MEETING_WEEK_2)
        self.should_not_conflict(
            sample_data.QUADWEEKLY_MEETING,
            sample_data.QUADWEEKLY_MEETING_WEEK_3)
        self.should_not_conflict(
            sample_data.QUADWEEKLY_MEETING_WEEK_1,
            sample_data.QUADWEEKLY_MEETING_WEEK_2)
        self.should_not_conflict(
            sample_data.QUADWEEKLY_MEETING_WEEK_1,
            sample_data.QUADWEEKLY_MEETING_WEEK_3)
        self.should_not_conflict(
            sample_data.QUADWEEKLY_MEETING_WEEK_2,
            sample_data.QUADWEEKLY_MEETING_WEEK_3)
        self.should_be_conflicting(
            sample_data.QUADWEEKLY_MEETING,
            sample_data.QUADWEEKLY_MEETING)
        self.should_be_conflicting(
            sample_data.QUADWEEKLY_MEETING_WEEK_1,
            sample_data.QUADWEEKLY_MEETING_WEEK_1)
        self.should_be_conflicting(
            sample_data.QUADWEEKLY_MEETING_WEEK_2,
            sample_data.QUADWEEKLY_MEETING_WEEK_2)
        self.should_be_conflicting(
            sample_data.QUADWEEKLY_MEETING_WEEK_3,
            sample_data.QUADWEEKLY_MEETING_WEEK_3)
        self.should_be_conflicting(
            sample_data.QUADWEEKLY_MEETING_WEEK_2,
            sample_data.QUADWEEKLY_MEETING_ALTERNATE)

    def test_meeting_duration(self):
        m = meeting.load_meetings(sample_data.MEETING_WITH_DURATION)[0]
        self.assertEqual(30, m.schedules[0].duration)
        m = meeting.load_meetings(sample_data.WEEKLY_MEETING)[0]
        self.assertEqual(60, m.schedules[0].duration)

    def test_short_meeting_conflicts(self):
        self.should_be_conflicting(
            sample_data.WEEKLY_MEETING,
            sample_data.MEETING_WITH_DURATION)
        self.should_not_conflict(
            sample_data.CONFLICTING_WEEKLY_MEETING,
            sample_data.MEETING_WITH_DURATION)

    def test_monthly_conflicts(self):
        self.should_be_conflicting(
            sample_data.WEEKLY_MEETING_2200,
            sample_data.FIRST_WEDNESDAY_MEETING)
        self.should_be_conflicting(
            sample_data.BIWEEKLY_EVEN_MEETING,
            sample_data.FIRST_WEDNESDAY_MEETING)
        self.should_be_conflicting(
            sample_data.QUADWEEKLY_MEETING,
            sample_data.FIRST_WEDNESDAY_MEETING)
        self.should_be_conflicting(
            sample_data.ALTERNATING_MEETING,
            sample_data.FIRST_WEDNESDAY_MEETING)

    def test_skip_meeting(self):
        meeting_yaml = sample_data.MEETING_WITH_SKIP_DATES
        # Copied from sample_data.MEETING_WITH_SKIP_DATES
        summary = 'OpenStack Subteam 8 Meeting'
        patterns = []
        # The "main" meeting should have an exdate
        patterns.append(re.compile(r'exdate:20150810T120000Z', re.I))
        # The "main" meeting should start on 2015-08-13
        patterns.append(re.compile(r'dtstart:20150803T120000Z', re.I))
        # The "main" meeting should have a simple summary
        patterns.append(re.compile(r'summary:%s' % summary, re.I))
        # The "skipped" meeting should start on 20150806
        patterns.append(re.compile(r'dtstart:20150810T120000Z', re.I))
        # The "skipped" meeting should say include 'CANCELLED' and the datetime
        patterns.append(re.compile(r'summary:CANCELLED.*20150810T120000Z',
                                   re.I))
        m = meeting.load_meetings(meeting_yaml)[0]
        cal = ical.Yaml2IcalCalendar()
        cal.add_meeting(m)
        cal_str = cal.to_ical().decode('utf-8')
        self.assertTrue(hasattr(m.schedules[0], 'skip_dates'))
        for p in patterns:
            self.assertIsNotNone(p.search(cal_str))

    def test_adhoc_meeting(self):
        meeting_yaml = sample_data.ADHOC_MEETING
        m = meeting.load_meetings(meeting_yaml)[0]
        cal = ical.Yaml2IcalCalendar()
        cal.add_meeting(m)
        self.assertEqual(cal, ical.Yaml2IcalCalendar())

    def test_alternating_chairs_meeting(self):
        meeting_yaml = sample_data.ALTERNATING_CHAIRS_MEETING
        m = meeting.load_meetings(meeting_yaml)[0]
        cal = ical.Yaml2IcalCalendar()
        cal.add_meeting(m)
        self.assertEqual(2, len(cal.events), cal)
        self.assertIn('Chair:  Jane Developer', cal.events[0]['description'])
        self.assertIn('Chair:  John Developer', cal.events[1]['description'])

    def test_skip_meeting_missing_skip_date(self):
        self.assertRaises(KeyError,
                          meeting.load_meetings,
                          sample_data.MEETING_WITH_MISSING_SKIP_DATE)

    def test_skip_meeting_missing_reason(self):
        self.assertRaises(KeyError,
                          meeting.load_meetings,
                          sample_data.MEETING_WITH_MISSING_REASON)

    def test_skip_meeting_bad_skip_date(self):
        self.assertRaises(ValueError,
                          meeting.load_meetings,
                          sample_data.MEETING_WITH_SKIP_DATES_BAD_DATE)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import unittest

from yaml2ical import recurrence


class RecurrenceTestCase(unittest.TestCase):

    def next_meeting(self, to_test):
        test_time = datetime.datetime(2014, 10, 5, 2, 47, 28, 832666)
        test_weekday = 'Wednesday'
        return to_test.next_occurence(test_time, test_weekday)

    def test_next_weekly(self):
        self.assertEqual(
            datetime.datetime(2014, 10, 8, 2, 47, 28, 832666),
            self.next_meeting(recurrence.WeeklyRecurrence()))

    def test_next_biweekly_odd(self):
        self.assertEqual(
            datetime.datetime(2014, 10, 8, 2, 47, 28, 832666),
            self.next_meeting(recurrence.BiWeeklyRecurrence(style='odd')))

    def test_next_biweekly_even(self):
        self.assertEqual(
            datetime.datetime(2014, 10, 15, 2, 47, 28, 832666),
            self.next_meeting(recurrence.BiWeeklyRecurrence(style='even')))

    def test_next_quadweekly_week_0(self):
        self.assertEqual(
            datetime.datetime(2014, 10, 29, 2, 47, 28, 832666),
            self.next_meeting(recurrence.QuadWeeklyRecurrence(week=0)))

    def test_next_quadweekly_week_1(self):
        self.assertEqual(
            datetime.datetime(2014, 10, 8, 2, 47, 28, 832666),
            self.next_meeting(recurrence.QuadWeeklyRecurrence(week=1)))

    def test_next_quadweekly_week_2(self):
        self.assertEqual(
            datetime.datetime(2014, 10, 15, 2, 47, 28, 832666),
            self.next_meeting(recurrence.QuadWeeklyRecurrence(week=2)))

    def test_next_quadweekly_week_3(self):
        self.assertEqual(
            datetime.datetime(2014, 10, 22, 2, 47, 28, 832666),
            self.next_meeting(recurrence.QuadWeeklyRecurrence(week=3)))

    def test_next_adhoc(self):
        self.assertEqual(
            None,
            self.next_meeting(recurrence.AdhocRecurrence()))

    def test_weekly_str(self):
        self.assertEqual(
            'Weekly',
            str(recurrence.WeeklyRecurrence()))

    def test_biweekly_odd_str(self):
        self.assertEqual(
            'Every two weeks (on odd weeks)',
            str(recurrence.BiWeeklyRecurrence(style='odd')))

    def test_biweekly_even_str(self):
        self.assertEqual(
            'Every two weeks (on even weeks)',
            str(recurrence.BiWeeklyRecurrence(style='even')))

    def test_quadweekly_str(self):
        for i in range(4):
            self.assertEqual(
                'Every four weeks on week %d of the four week rotation' % i,
                str(recurrence.QuadWeeklyRecurrence(week=i)))

    def test_monthly_first_week(self):
        rec = recurrence.MonthlyRecurrence(week=1, day='Wednesday')
        self.assertEqual(
            datetime.datetime(2014, 11, 5, 2, 47, 28, 832666),
            self.next_meeting(rec),
        )

    def test_monthly_second_week(self):
        rec = recurrence.MonthlyRecurrence(week=2, day='Wednesday')
        self.assertEqual(
            datetime.datetime(2014, 11, 12, 2, 47, 28, 832666),
            self.next_meeting(rec),
        )

    def test_monthly_third_week(self):
        rec = recurrence.MonthlyRecurrence(week=3, day='Thursday')
        self.assertEqual(
            datetime.datetime(2014, 11, 19, 2, 47, 28, 832666),
            self.next_meeting(rec),
        )

    def test_monthly_fourth_week(self):
        rec = recurrence.MonthlyRecurrence(week=4, day='Tuesday')
        self.assertEqual(
            datetime.datetime(2014, 11, 26, 2, 47, 28, 832666),
            self.next_meeting(rec),
        )

    def test_monthly_invalid_week(self):
        rec = recurrence.MonthlyRecurrence(week=6, day='Wednesday')
        self.assertRaises(
            ValueError,
            self.next_meeting,
            rec,
        )

    def test_monthly_day_specifier(self):
        weeks = [
            (1, 'the first'),
            (2, 'the second'),
            (3, 'the third'),
            (4, 'the fourth'),
            (5, 'the fifth'),
        ]
        for i, expected in weeks:
            rec = recurrence.MonthlyRecurrence(week=i, day='Thursday')
            self.assertEqual(expected, rec.day_specifier)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import logging
import os

from yaml2ical import ical
from yaml2ical import index
from yaml2ical import meeting


# logging settings
logging.basicConfig(format='%(asctime)s  - %(levelname)s - %(message)s',
                    level=logging.DEBUG)


def parse_args():
    # build option parser:
    description = """
A tool that automates the process for testing, integrating, and
publishing changes to online meeting schedules.
"""

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=description)

    parser.add_argument("-y", "--yamldir",
                        dest="yaml_dir",
                        required=True,
                        help="directory containing YAML to process")
    outputtype = parser.add_mutually_exclusive_group(required=True)
    outputtype.add_argument("-i", "--icaldir",
                            dest="ical_dir",
                            help="output directory (one file per meeting)")
    outputtype.add_argument("-o", "--output",
                            dest="icalfile",
                            help="output file (one file for all meetings)")
    parser.add_argument("-t", "--indextemplate",
                        dest="index_template",
                        help="generate an index from selected meetings")
    parser.add_argument("-w", "--indexoutput",
                        dest="index_output",
                        help="output index file")
    parser.add_argument("-n", "--name",
                        dest="calname",
                        help="name of calendar to set within the ical")
    parser.add_argument("-d", "--description",
                        dest="caldescription",
                        help="description of calendar to set within the ical")
    parser.add_argument("-f", "--force",
                        dest="force",
                        action='store_true',
                        help="remove/overwrite previous output files")

    # parse arguments:
    args = parser.parse_args()

    if (
            (args.index_template and not args.index_output)
            or (args.index_output and not args.index_template)):
        parser.error("You need to provide both -t and "
                     "-w if you want to output an index.")
    if args.ical_dir and (args.calname or args.caldescription):
        parser.error("Name/Description and single ical per meeting "
                     "(-i) is incompatiable due to spec.")

    return args


def _check_if_location_exists(location, style='f'):
    if style == 'd' and not os.path.isdir(location):
        raise ValueError("Directory %s does not exist" % location)
    if style == 'f' and not os.path.isfile(location):
        raise ValueError("File %s does not exist" % location)


def _prepare_output(output, style='f', force=False):
    location = os.path.abspath(output)
    if style == 'd':
        _check_if_location_exists(location, style=style)
        if os.listdir(location) != []:
            if force:
                for f in os.listdir(location):
                    file_path = os.path.join(location, f)
                    os.remove(file_path)
            else:
                raise Exception("Directory for storing iCals is not empty, "
                                "suggest running with -f to remove old files.")
    else:
        if os.path.exists(location):
            if force:
                os.remove(location)
            else:
                raise Exception("Output file already exists, suggest running "
                                "with -f to overwrite previous file.")
    return location


def main():
    args = parse_args()

    yaml_dir = os.path.abspath(args.yaml_dir)
    _check_if_location_exists(yaml_dir, style='d')

    meetings = meeting.load_meetings(yaml_dir)
    # Check uniqueness and conflicts here before writing out to .ics
    meeting.check_for_meeting_conflicts(meetings)

    if args.ical_dir:
        ical_dir = _prepare_output(args.ical_dir, style='d', force=args.force)
        ical.convert_meetings_to_ical(meetings, outputdir=ical_dir)
    else:
        icalfile = _prepare_output(args.icalfile, force=args.force)
        ical.convert_meetings_to_ical(meetings, outputfile=icalfile,
                                      caldescription=args.caldescription,
                                      calname=args.calname)

    if args.index_template and args.index_output:
        index_template = os.path.abspath(args.index_template)
        _check_if_location_exists(index_template)
        index_output = _prepare_output(args.index_output, force=args.force)
        index.convert_meetings_to_index(
            meetings, index_template, index_output)


if __name__ == '__main__':
    main()



# Copyright 2016 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest

from yaml2ical import index


class IndexTestCase(unittest.TestCase):

    def setUp(self):
        # Make 23 entries as 23 is a prime number
        self.meetings = [x for x in range(1, 24)]

    def test_batch_meetings_3_cols(self):
        # NOTE(jlvillal): Please format the expected values to easily see the
        # columns and rows
        tests = (
            # 1 row
            (1, [1]),
            (2, [1, 2]),
            (3, [1, 2, 3]),
            # 2 rows
            (4, [1, 3, 4,
                 2]),
            (5, [1, 3, 5,
                 2, 4]),
            # 3 rows
            (7, [1, 4, 6,
                 2, 5, 7,
                 3]),
            (9, [1, 4, 7,
                 2, 5, 8,
                 3, 6, 9]),
            # 4 rows
            (11, [1, 5, 9,
                  2, 6, 10,
                  3, 7, 11,
                  4, 8]),
            # 8 rows
            (23, [1, 9, 17,
                  2, 10, 18,
                  3, 11, 19,
                  4, 12, 20,
                  5, 13, 21,
                  6, 14, 22,
                  7, 15, 23,
                  8, 16]),
        )

        for test_length, expected in tests:
            self.assertEqual(
                expected, index.batch_meetings(self.meetings[:test_length], 3))

    def test_batch_meetings_4_cols(self):
        # NOTE(jlvillal): Please format the expected values to easily see the
        # columns and rows
        tests = (
            # Empty
            (0, []),
            # 1 row
            (1, [1]),
            (2, [1, 2]),
            (3, [1, 2, 3]),
            (4, [1, 2, 3, 4]),
            # 2 rows
            (5, [1, 3, 4, 5,
                 2]),
            (7, [1, 3, 5, 7,
                 2, 4, 6]),
            (8, [1, 3, 5, 7,
                 2, 4, 6, 8]),
            # 3 rows
            (9, [1, 4, 6, 8,
                 2, 5, 7, 9,
                 3]),
            (11, [1, 4, 7, 10,
                  2, 5, 8, 11,
                  3, 6, 9]),
            # 5 rows
            (23, [1, 7, 13, 19,
                  2, 8, 14, 20,
                  3, 9, 15, 21,
                  4, 10, 16, 22,
                  5, 11, 17, 23,
                  6, 12, 18]),
        )

        for test_length, expected in tests:
            self.assertEqual(
                expected, index.batch_meetings(self.meetings[:test_length], 4))

        # Make sure our docstring example is correct
        self.assertEqual(
            ['A', 'D', 'F', 'H',
             'B', 'E', 'G', 'I',
             'C'],
            index.batch_meetings(['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I'],
                                 4))

    def test_batch_meetings_misc_cols(self):
        # Test various column values and inputs

        # NOTE(jlvillal): Please format the expected values to easily see the
        # columns and rows
        tests = (
            # Formatted as:
            # Number of items, columns, expected_result

            # Empty
            (0, 1, []),
            # 2 rows
            (8, 5, [1, 3, 5, 7, 8,
                    2, 4, 6]),
            (10, 6, [1, 3, 5, 7, 9, 10,
                     2, 4, 6, 8, ]),
            (7, 6, [1, 3, 4, 5, 6, 7,
                    2]),
        )

        for test_length, columns, expected in tests:
            self.assertEqual(
                expected,
                index.batch_meetings(self.meetings[:test_length], columns))

    def test_batch_meetings_zero_or_less(self):
        # Make sure we return the passed in value if columns less than or equal
        # to zero
        self.assertEqual([1, 2], index.batch_meetings([1, 2], 0))
        self.assertEqual([1, 2], index.batch_meetings([1, 2], -1))
        self.assertEqual([1, 2], index.batch_meetings([1, 2], -5551.5))

    def test_batch_meetings_sanity_check_not_triggered(self):
        # Make sure that an input containing None, False, 0, or "" doesn't
        # trigger the assert
        self.assertEqual([None, None], index.batch_meetings([None, None], 1))
        self.assertEqual(
            [False, False], index.batch_meetings([False, False], 1))
        self.assertEqual([0, 0], index.batch_meetings([0, 0], 4))
        self.assertEqual(["", ""], index.batch_meetings(["", ""], 1))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import unittest

from yaml2ical import meeting
from yaml2ical import recurrence
from yaml2ical.tests import sample_data


class Meeting_RecurrenceTestCase(unittest.TestCase):

    def test_next_meeting_start_date(self):
        m = meeting.load_meetings(sample_data.MEETING_WITH_START_DATE)[0]
        self.assertEqual(
            datetime.datetime(2015, 8, 6, 0, 0),
            recurrence.WeeklyRecurrence().next_occurence(
                m.schedules[0].start_date, m.schedules[0].day))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import logging
import os
import os.path

import icalendar
import pytz


class Yaml2IcalCalendar(icalendar.Calendar):
    """A calendar in ics format."""

    def __init__(self, calname=None, caldescription=None):
        super(Yaml2IcalCalendar, self).__init__()
        self.add('prodid', '-//yaml2ical agendas//EN')
        self.add('version', '2.0')
        if calname is not None:
            self.add('X-WR-CALNAME', calname)
        if caldescription is not None:
            self.add('X-WR-CALDESC', caldescription)

    def add_meeting(self, meeting):
        """Add this meeting to the calendar."""

        for sch in meeting.schedules:
            self.add_schedule(meeting, sch)

    def add_schedule(self, meeting, sch, exdate=None):
        event = icalendar.Event()

        # NOTE(jotan): I think the summary field needs to be unique per
        # event in an ical file (at least, for it to work with
        # Google Calendar)

        summary = meeting.project
        # NOTE(tonyb): If we're adding an a place holder event for a
        # cancelled meeting make that as obvious as possible in the
        # summary.
        if exdate:
            # NOTE(tonyb): Because some iCal consumers require that the
            # summary be unique, and adding multiple "CANCELLED: $x"
            # entries would violate that rule, append the (UTC)
            # timestamp for the cancelled meeting.
            suffix = exdate.date_str
            summary = 'CANCELLED: %s (%s)' % (summary, suffix)
        event.add('summary', summary)
        event.add('location', '#' + sch.irc)

        # add ical description
        project_descript = "Project:  %s" % (meeting.project)
        chair_descript = "Chair:  %s" % (sch.chair)
        descript_descript = "Description:  %s" % (meeting.description)
        ical_descript = "\n".join((project_descript,
                                   chair_descript,
                                   descript_descript))
        # Add URLs, if present, to the description
        if 'agenda_url' in meeting.extras:
            ical_descript = "\n".join((ical_descript,
                                       "Agenda URL:  %s" %
                                       (meeting.extras['agenda_url'])))
        if 'project_url' in meeting.extras:
            ical_descript = "\n".join((ical_descript,
                                       "Project URL:  %s" %
                                       (meeting.extras['project_url'])))

        # NOTE(tonyb): If we're adding an a place holder event for a
        # cancelled meeting do not add an rrule and set dtstart to the
        # skipped date.
        if not exdate:
            # get starting date
            next_meeting = sch.recurrence.next_occurence(sch.start_date,
                                                         sch.day)
            # NOTE(aschultz): to handle adhoc meetings, we check to make
            # sure there is a next meeting before trying to add it to the
            # the calendar
            if next_meeting is None:
                return
            next_meeting_date = datetime.datetime(next_meeting.year,
                                                  next_meeting.month,
                                                  next_meeting.day,
                                                  sch.time.hour,
                                                  sch.time.minute,
                                                  tzinfo=pytz.utc)
            event.add('dtstart', next_meeting_date)

            # add recurrence rule
            event.add('rrule', sch.recurrence.rrule())
            event.add('description', ical_descript)
        else:
            event.add('dtstart', exdate.date)
            event.add('description', exdate.reason)

        event.add('duration', datetime.timedelta(minutes=sch.duration))

        # Add exdate (exclude date) if present
        if not exdate and hasattr(sch, 'skip_dates'):
            for skip_date in sch.skip_dates:
                event.add('exdate', skip_date.date)
                # NOTE(tonyb): If this is a skipped meeting add a
                # non-recurring event with an obvious summary.
                self.add_schedule(meeting, sch, exdate=skip_date)

        # Add update timestamp and unique ID for the event.
        # Note: we need a way to uniquely identify the meeting, even if its
        # details are modified. We don't really have much to go on to do
        # that, so best effort just use the project name and date.
        event.add('dtstamp', meeting.last_update)
        start_date = exdate.date if exdate else next_meeting_date
        event.add('uid', '%s-%s' % (
                  meeting.project.replace(' ', '').lower(),
                  datetime.datetime.strftime(start_date, '%Y%m%d')))

        # add event to calendar
        self.add_component(event)

    def write_to_disk(self, filename):
        # write ical files to disk
        with open(filename, 'wb') as ics:
            ics.write(self.to_ical())


def convert_meetings_to_ical(meetings, outputdir=None, outputfile=None,
                             calname=None, caldescription=None):
    """Converts a meeting list to iCal.

    :param meetings: list of meetings to convert
    :param outputdir: location to store iCal files (one file per meeting)
    :param outputfile: output iCal file (one single file for all meetings)

    """

    # convert meetings to a list of ical
    if outputdir:
        for m in meetings:
            cal = Yaml2IcalCalendar()
            cal.add_meeting(m)
            cal.write_to_disk(os.path.join(outputdir, m.outfile))

    # convert meetings into a single ical
    if outputfile:
        cal = Yaml2IcalCalendar(calname, caldescription)
        for m in meetings:
            cal.add_meeting(m)
        cal.write_to_disk(outputfile)

    # TODO(jotan): verify converted ical is valid
    logging.info('Wrote %d meetings to iCal' % (len(meetings)))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import calendar
import datetime


WEEKDAYS = {'Monday': 0, 'Tuesday': 1, 'Wednesday': 2, 'Thursday': 3,
            'Friday': 4, 'Saturday': 5, 'Sunday': 6}


class _Recurrence(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def next_occurence(self, current_date_time, day):
        """Return the date of the next meeting.

        :param ref_date: datetime object of meeting
        :param day: weekday the meeting is held on

        :returns: datetime object of the next meeting time
        """

    @abc.abstractmethod
    def rrule(self):
        "Return dict with data for building ICS recurrence rule."

    @abc.abstractmethod
    def __str__(self):
        "Return string representation of the recurrence rule"

    @property
    def day_specifier(self):
        """Return string prefix for day.

        For example, monthly recurring events may return 'first' to
        indicate the first instance of a particular day of the month
        (e.g., first Thursday).
        """
        # NOTE(dhellmann): This is not an abstract property because
        # most of the subclasses will use this concrete
        # implementation.
        return ''


class WeeklyRecurrence(_Recurrence):
    """Meetings occuring every week."""
    def __init__(self):
        pass

    def next_occurence(self, current_date_time, day):
        """Return the date of the next meeting.

        :param ref_date: datetime object of meeting
        :param day: weekday the meeting is held on

        :returns: datetime object of the next meeting time
        """

        weekday = WEEKDAYS[day]
        days_ahead = weekday - current_date_time.weekday()
        if days_ahead < 0:  # target day already happened this week
            days_ahead += 7
        return current_date_time + datetime.timedelta(days_ahead)

    def rrule(self):
        return {'freq': 'weekly'}

    def __str__(self):
        return "Weekly"


class BiWeeklyRecurrence(_Recurrence):
    """Meetings occuring on alternate weeks.

    Can be either on odd weeks or on even weeks
    """
    def __init__(self, style='even'):
        self.style = style

    def next_occurence(self, current_date, day):
        """Calculate the next biweekly meeting.

        :param current_date: the current date
        :param day: scheduled day of the meeting
        :returns: datetime object of next meeting
        """
        nextweek_day = WeeklyRecurrence().next_occurence(current_date, day)
        if nextweek_day.isocalendar()[1] % 2:
            # ISO week is odd
            if self.style == 'odd':
                return nextweek_day
        else:
            # ISO week is even
            if self.style == 'even':
                return nextweek_day
        # If week doesn't match rule, skip one week
        return nextweek_day + datetime.timedelta(7)

    def rrule(self):
        return {'freq': 'weekly', 'interval': 2}

    def __str__(self):
        return "Every two weeks (on %s weeks)" % self.style


class QuadWeeklyRecurrence(_Recurrence):
    """Meetings occuring every 4 weeks.

    A week number can be supplied to offset meetings
    """
    def __init__(self, week=0):
        self.week = week

    def next_occurence(self, current_date, day):
        """Calculate the next biweekly meeting.

        :param current_date: the current date
        :param day: scheduled day of the meeting
        :returns: datetime object of next meeting
        """
        nextweek_day = WeeklyRecurrence().next_occurence(current_date, day)
        if nextweek_day.isocalendar()[1] % 4 == self.week:
            return nextweek_day
        # If week doesn't match rule, skip one week
        return self.next_occurence(nextweek_day + datetime.timedelta(7), day)

    def rrule(self):
        return {'freq': 'weekly', 'interval': 4}

    def __str__(self):
        return (
            "Every four weeks on week %d of the four week rotation"
            % self.week)


class AdhocRecurrence(_Recurrence):
    """Meetings occuring as needed.

    Effectively this is a noop recurrance as next_occurance is always None
    """
    def __init__(self):
        pass

    def next_occurence(self, current_date, day):
        """Calculate the next adhoc meeting.

        :param current_date: the current date
        :param day: scheduled day of the meeting
        :returns: datetime object of next meeting
        """
        return None

    def rrule(self):
        return {'freq': 'adhoc', 'interval': 0}

    def __str__(self):
        return "Occurs as needed, no fixed schedule."


class MonthlyRecurrence(_Recurrence):
    """Meetings occuring every month."""
    def __init__(self, week, day):
        self._week = week
        self._day = day

    def next_occurence(self, current_date_time, day):
        """Return the date of the next meeting.

        :param current_date_time: datetime object of meeting
        :param day: weekday the meeting is held on

        :returns: datetime object of the next meeting time
        """
        weekday = WEEKDAYS[day]

        month = current_date_time.month + 1
        year = current_date_time.year
        if current_date_time.month == 12:
            month = 1
            year = year + 1
        next_month_dates = calendar.monthcalendar(year, month)

        # We can't simply index into the dates for the next month
        # because we don't know that the first week is full of days
        # that actually appear in that month. Therefore we loop
        # through them counting down until we've skipped enough weeks.
        skip_weeks = self._week - 1
        for week in next_month_dates:
            day = week[weekday]
            # Dates in the week that fall in other months
            # are 0 so we want to skip counting those weeks.
            if not day:
                continue
            # If we have skipped all of the weeks we need to,
            # we have the day.
            if not skip_weeks:
                return datetime.datetime(
                    year, month, day,
                    current_date_time.hour, current_date_time.minute,
                    current_date_time.second, current_date_time.microsecond,
                )
            skip_weeks -= 1

        raise ValueError(
            'Could not compute week {} of next month for {}'.format(
                self._week, current_date_time)
        )

    def rrule(self):
        return {
            'freq': 'monthly',
            'byday': '{}{}'.format(self._week, self._day[:2].upper()),
        }

    _ORDINALS = [
        'first',
        'second',
        'third',
        'fourth',
        'fifth',
    ]

    @property
    def day_specifier(self):
        return 'the {}'.format(self._ORDINALS[self._week - 1])

    def __str__(self):
        return "Monthly, in the " + format(
            self._ORDINALS[self._week - 1]) + ' week,'


supported_recurrences = {
    'weekly': WeeklyRecurrence(),
    'biweekly-odd': BiWeeklyRecurrence(style='odd'),
    'biweekly-even': BiWeeklyRecurrence(),
    'quadweekly': QuadWeeklyRecurrence(week=0),
    'quadweekly-week-1': QuadWeeklyRecurrence(week=1),
    'quadweekly-week-2': QuadWeeklyRecurrence(week=2),
    'quadweekly-week-3': QuadWeeklyRecurrence(week=3),
    'quadweekly-alternate': QuadWeeklyRecurrence(week=2),
    'adhoc': AdhocRecurrence(),
    'first-monday': MonthlyRecurrence(week=1, day='Monday'),
    'first-tuesday': MonthlyRecurrence(week=1, day='Tuesday'),
    'first-wednesday': MonthlyRecurrence(week=1, day='Wednesday'),
    'first-thursday': MonthlyRecurrence(week=1, day='Thursday'),
    'first-friday': MonthlyRecurrence(week=1, day='Friday'),
    'second-monday': MonthlyRecurrence(week=2, day='Monday'),
    'second-tuesday': MonthlyRecurrence(week=2, day='Tuesday'),
    'second-wednesday': MonthlyRecurrence(week=2, day='Wednesday'),
    'second-thursday': MonthlyRecurrence(week=2, day='Thursday'),
    'second-friday': MonthlyRecurrence(week=2, day='Friday'),
    'third-monday': MonthlyRecurrence(week=3, day='Monday'),
    'third-tuesday': MonthlyRecurrence(week=3, day='Tuesday'),
    'third-wednesday': MonthlyRecurrence(week=3, day='Wednesday'),
    'third-thursday': MonthlyRecurrence(week=3, day='Thursday'),
    'third-friday': MonthlyRecurrence(week=3, day='Friday'),
    'fourth-monday': MonthlyRecurrence(week=4, day='Monday'),
    'fourth-tuesday': MonthlyRecurrence(week=4, day='Tuesday'),
    'fourth-wednesday': MonthlyRecurrence(week=4, day='Wednesday'),
    'fourth-thursday': MonthlyRecurrence(week=4, day='Thursday'),
    'fourth-friday': MonthlyRecurrence(week=4, day='Friday'),
}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
from io import StringIO
import os
import os.path
import pytz
import subprocess
import yaml


from yaml2ical.recurrence import supported_recurrences

DATES = {
    'Monday': datetime.date(1900, 1, 1),
    'Tuesday': datetime.date(1900, 1, 2),
    'Wednesday': datetime.date(1900, 1, 3),
    'Thursday': datetime.date(1900, 1, 4),
    'Friday': datetime.date(1900, 1, 5),
    'Saturday': datetime.date(1900, 1, 6),
    'Sunday': datetime.date(1900, 1, 7),
}
ONE_WEEK = datetime.timedelta(weeks=1)


class SkipDate(object):
    """A date, time and reason to skip a meeting."""

    def __init__(self, date, time, reason):
        date = datetime.datetime.combine(date, time).replace(tzinfo=pytz.utc)
        self.date = date
        self.reason = reason

    @property
    def date_str(self):
        return self.date.strftime("%Y%m%dT%H%M%SZ")


class Schedule(object):
    """A meeting schedule."""

    def __init__(self, meeting, sched_yaml):
        """Initialize schedule from yaml."""

        self.project = meeting.project
        self.filefrom = meeting.filefrom
        self.chair = sched_yaml.get('chair', meeting.chair)
        # mandatory: time, day, irc, freq, recurrence
        try:
            self.utc = sched_yaml['time']
            self.time = datetime.datetime.strptime(sched_yaml['time'], '%H%M')
            # Sanitize the Day
            self.day = sched_yaml['day'].lower().capitalize()
            self.irc = sched_yaml['irc']
            self.freq = sched_yaml['frequency']
            self._recurrence = sched_yaml['frequency']
        except KeyError as e:
            print("Invalid YAML meeting schedule definition - missing "
                  "attribute '{0}'".format(e.args[0]))
            raise

        # Validate inputs
        try:
            self.recurrence = supported_recurrences[self._recurrence]
        except KeyError as e:
            print("Invalid meeting recurrence '{0}' - "
                  "valid types: {1}".format(e.args[0],
                                            supported_recurrences.keys()))
            raise

        # optional: start_date defaults to the current date if not present
        if 'start_date' in sched_yaml:
            try:
                self.start_date = datetime.datetime.strptime(
                    str(sched_yaml['start_date']), '%Y%m%d')
            except ValueError:
                raise ValueError("Could not parse 'start_date' (%s) in %s" %
                                 (sched_yaml['start_date'], self.filefrom))
        else:
            self.start_date = datetime.datetime.utcnow()

        # optional: duration
        if 'duration' in sched_yaml:
            try:
                self.duration = int(sched_yaml['duration'])
            except ValueError:
                raise ValueError("Could not parse 'duration' (%s) in %s" %
                                 (sched_yaml['duration'], self.filefrom))
        else:
            self.duration = 60

        if self.day not in DATES.keys():
            raise ValueError("'%s' is not a valid day of the week")

        # optional: skip_dates
        # This is a sequence of mappings (YAML)
        # This is a list of dicts (python)
        if 'skip_dates' in sched_yaml:
            self.skip_dates = []
            for skip_date in sched_yaml['skip_dates']:
                missing_keys = set(['skip_date', 'reason']) - set(skip_date)
                if missing_keys:
                    raise KeyError(("Processing: %s Missing keys - %s" %
                                    (self.filefrom, ','.join(missing_keys))))

                # NOTE(tonyb) We need to include the time in an exdate
                # without it the excluded occurrence never matches a
                # scheduled occurrence.
                try:
                    date_str = str(skip_date['skip_date'])
                    date = datetime.datetime.strptime(date_str, '%Y%m%d')
                    self.skip_dates.append(SkipDate(date, self.time.time(),
                                                    skip_date['reason']))
                except ValueError:
                    raise ValueError(("Processing: %s Could not parse "
                                      "skip_date - %s" %
                                      (self.filefrom, skip_date['skip_date'])))

        # NOTE(tonyb): We need to do this datetime shenanigans is so we can
        #              deal with meetings that start on day1 and end on day2.
        self.meeting_start = datetime.datetime.combine(DATES[self.day],
                                                       self.time.time())
        self.meeting_end = (self.meeting_start
                            + datetime.timedelta(minutes=self.duration))
        if self.day == 'Sunday' and self.meeting_end.strftime("%a") == 'Mon':
            self.meeting_start = self.meeting_start - ONE_WEEK
            self.meeting_end = self.meeting_end - ONE_WEEK

    def conflicts(self, other):
        """Checks for conflicting schedules."""

        def _non_weekly_conflict_detection(self, other):
            week = {
                'weekly': set([0, 1, 2, 3]),
                'biweekly-even': set([0, 2]),
                'biweekly-odd': set([1, 3]),
                'quadweekly': set([0]),
                'quadweekly-week-1': set([1]),
                'quadweekly-week-2': set([2]),
                'quadweekly-week-3': set([3]),
                'quadweekly-alternate': set([2]),
                'first-monday': set([0, 1, 2, 3]),
                'first-tuesday': set([0, 1, 2, 3]),
                'first-wednesday': set([0, 1, 2, 3]),
                'first-thursday': set([0, 1, 2, 3]),
                'first-friday': set([0, 1, 2, 3]),
                'second-monday': set([0, 1, 2, 3]),
                'second-tuesday': set([0, 1, 2, 3]),
                'second-wednesday': set([0, 1, 2, 3]),
                'second-thursday': set([0, 1, 2, 3]),
                'second-friday': set([0, 1, 2, 3]),
                'third-monday': set([0, 1, 2, 3]),
                'third-tuesday': set([0, 1, 2, 3]),
                'third-wednesday': set([0, 1, 2, 3]),
                'third-thursday': set([0, 1, 2, 3]),
                'third-friday': set([0, 1, 2, 3]),
                'fourth-monday': set([0, 1, 2, 3]),
                'fourth-tuesday': set([0, 1, 2, 3]),
                'fourth-wednesday': set([0, 1, 2, 3]),
                'fourth-thursday': set([0, 1, 2, 3]),
                'fourth-friday': set([0, 1, 2, 3]),
            }

            return len(week[self.freq].intersection(week[other.freq])) > 0

        # NOTE(tonyb): .meeting_start also includes the day of the week. So no
        #              need to check .day explictly
        return ((self.irc == other.irc)
                and ((self.meeting_start < other.meeting_end)
                     and (other.meeting_start < self.meeting_end))
                and _non_weekly_conflict_detection(self, other))


class Meeting(object):
    """An online meeting."""

    def __init__(self, data):
        """Initialize meeting from meeting yaml description."""

        yaml_obj = yaml.safe_load(data)

        try:
            self.description = yaml_obj['description']
            self.project = yaml_obj['project']
        except KeyError as e:
            print("Invalid YAML meeting definition - missing "
                  "attribute '{0}'".format(e.args[0]))
            raise

        try:
            self.chair = yaml_obj['chair']
        except KeyError as e:
            if any('chair' not in s for s in yaml_obj['schedule']):
                print("Invalid YAML meeting definition - missing "
                      "attribute '{0}'".format(e.args[0]))
                raise
            self.chair = None

        # Find any extra values the user has provided that they might
        # want to have access to in their templates.
        self.extras = {}
        self.extras.update(yaml_obj)
        for k in ['chair', 'description', 'project', 'schedule']:
            if k in self.extras:
                del self.extras[k]

        try:
            self.filefrom = os.path.basename(data.name)
            self.outfile = os.path.splitext(self.filefrom)[0] + '.ics'
        except AttributeError:
            self.filefrom = "stdin"
            self.outfile = "stdin.ics"

        self.schedules = []
        for sch in yaml_obj['schedule']:
            s = Schedule(self, sch)
            self.schedules.append(s)

        # Set a default last update time
        self.last_update = datetime.datetime.utcnow()

    @classmethod
    def fromfile(cls, yaml_file):
        with open(yaml_file, 'r') as f:
            meeting = cls(f)

        meeting.last_update = _get_update_time(yaml_file)
        return meeting

    @classmethod
    def fromstring(cls, yaml_string):
        s = StringIO(yaml_string)
        return cls(s)


def _get_update_time(src_file):
    """Attempts to figure out the last time the file was updated.

    If the file is under source control, this will try to find the last time it
    was updated. If that is not possible, it will fallback to using the last
    time the local file was modified.
    """
    try:
        last_updated = subprocess.check_output(
            [
                'git', 'log', '-n1', '--format=%ad',
                '--date=format:%Y-%m-%d %H:%M:%S',
                '--', src_file,
            ]
        ).decode('utf-8').strip()
        return datetime.datetime.strptime(last_updated, '%Y-%m-%d %H:%M:%S')
    except (subprocess.CalledProcessError, ValueError):
        return datetime.datetime.fromtimestamp(os.path.getmtime(src_file))


def load_meetings(yaml_source):
    """Build YAML object and load meeting data

    :param yaml_source: source data to load, which can be a directory or
                        stream.
    :returns: list of meeting objects
    """
    meetings = []
    # Determine what the yaml_source is. Files must have .yaml extension
    # to be considered valid.
    if os.path.isdir(yaml_source):
        for root, dirs, files in os.walk(yaml_source):
            for f in files:
                # Build the entire file path and append to the list of yaml
                # meetings
                if os.path.splitext(f)[1] == '.yaml':
                    yaml_file = os.path.join(root, f)
                    meetings.append(Meeting.fromfile(yaml_file))
    elif (os.path.isfile(yaml_source)
          and os.path.splitext(yaml_source)[1] == '.yaml'):
        meetings.append(Meeting.fromfile(yaml_source))
    elif isinstance(yaml_source, str):
        return [Meeting.fromstring(yaml_source)]

    if not meetings:
        # If we don't have a .yaml file, a directory of .yaml files, or any
        # YAML data fail out here.
        raise ValueError("No .yaml file, directory containing .yaml files, "
                         "or YAML data found.")
    else:
        meetings.sort(key=lambda x: x.project)
        return meetings


class MeetingConflictError(Exception):
    pass


def check_for_meeting_conflicts(meetings):
    """Check if a list of meetings have conflicts.

    :param meetings: list of Meeting objects

    """

    for i in range(len(meetings)):
        schedules = meetings[i].schedules
        for j in range(i + 1, len(meetings)):
            other_schedules = meetings[j].schedules
            for schedule in schedules:
                for other_schedule in other_schedules:
                    if schedule.conflicts(other_schedule):
                        msg_dict = {'one': schedule.filefrom,
                                    'two': other_schedule.filefrom}
                        raise MeetingConflictError(
                            "Conflict between %(one)s and %(two)s" % msg_dict)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import division

import datetime
import jinja2
import logging
import os
import os.path


def batch_meetings(meetings, batch_size):
    """Batches the meetings to be consumed by the jinja2 'batch' filter.

    This will pivot the meeting list into a virtual number of columns.  This
    can be used in a jinja template like:

    {% for column in batch_meetings(meetings, 4)|batch(4) %}

    So the list:
    [A, B, C, D, E, F , G, H, I]

    Is returned as:
    [A, D, F, H,   B, E, G, I,   C]]

    Or another way of looking at it is:
    [A, D, F, H,
     B, E, G, I,
     C]]

    And displays as:
    A D F H
    B E G I
    C

    Rather than:
    A B C D
    E F G H
    I

    :param meetings: An iterable
    :param batch_size: Number of columns to split up the output into
    :returns: A list that has 'pivoted' the meetings input

    """
    if batch_size <= 0:
        return meetings
    col_length = (len(meetings) // batch_size) + 1
    new_meetings = [None] * len(meetings)
    src = 0

    for row in range(batch_size):
        for col in range(col_length):
            dest = col * batch_size + row
            if dest >= len(meetings):
                break
            new_meetings[dest] = meetings[src]
            src += 1

    # Sanity check
    if all(meetings):
        assert all(new_meetings), "Empty item found in: {}".format(
            new_meetings)
    return new_meetings


def convert_meetings_to_index(meetings, template, output_file):
    """Creates index file from list of meetings.

    :param meetings: list of meetings to convert
    :param template: jinja2 template to use
    :param output_file: output index file

    """

    (template_dir, template_file) = os.path.split(template)
    loader = jinja2.FileSystemLoader(template_dir)
    env = jinja2.environment.Environment(trim_blocks=True, loader=loader)
    template = env.get_template(template_file)
    template.globals['batch_meetings'] = batch_meetings

    with open(output_file, "w") as out:
        out.write(template.render(meetings=meetings,
                                  timestamp=datetime.datetime.utcnow()))

    logging.info('Wrote %d meetings to index.' % (len(meetings)))



import nox


nox.options.error_on_external_run = True
nox.options.reuse_existing_virtualenvs = True
nox.options.sessions = ["tests-3", "linters"]


# Convenience wrapper for running the project
@nox.session(python="3")
def mkical(session):
    session.install(".")
    session.run("yaml2ical", "-y", "meetings/", "-i", "icals/", "-f")


# Note setting python this way seems to give us a target name without
# python specific suffixes while still allowing us to force a specific
# version using --force-python.
@nox.session(python="3")
def linters(session):
    # TODO: switch this line to 'session.install("--group", "test-linters")'
    session.install(".[test-linters]")
    session.run("flake8")


@nox.session(python="3")
def venv(session):
    # TODO: switch to 'session.install("-e", ".", "--group", "test-unit")'
    session.install("-e", ".[test-unit]")
    session.run(*session.posargs)


# This will attempt to run python3 tests by default.
@nox.session(python=["3"])
def tests(session):
    # TODO: switch to 'session.install("-e", ".", "--group", "test-unit")'
    session.install("-e", ".[test-unit]")
    session.run("stestr", "run", *session.posargs)
    session.run("stestr", "slowest")



#! /usr/bin/env python
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Sample meeting data to use for testing."""

BAD_MEETING_DAY = """
project: OpenStack Subteam Meeting
schedule:
  - time: '1200'
    day: go_bang
    irc: openstack-meeting
    frequency: weekly
chair: Joe Developer
description: >
    Weekly meeting for Subteam project.
"""

WEEKLY_MEETING = """
project: OpenStack Subteam Meeting
schedule:
  - time: '1200'
    day: Wednesday
    irc: openstack-meeting
    frequency: weekly
chair: Joe Developer
description: >
    Weekly meeting for Subteam project.
agenda: |
  * Top bugs this week
"""

CONFLICTING_WEEKLY_MEETING = """
project: OpenStack Subteam Meeting 2
schedule:
  - time: '1230'
    day: Wednesday
    irc: openstack-meeting
    frequency: weekly
chair: Joe Developer
description: >
    Weekly meeting for Subteam 2 project.
agenda: |
  * New features
"""

WEEKLY_OTHER_CHANNEL_MEETING = """
project: OpenStack Subteam Meeting 3
schedule:
  - time: '1200'
    day: Wednesday
    irc: openstack-meeting-alt
    frequency: weekly
chair: Joe Developer
description: >
    Weekly meeting for Subteam 3 project.
agenda: |
  * New features
"""

ALTERNATING_MEETING = """
project: OpenStack Subteam Meeting
schedule:
  - time: '1200'
    day: Wednesday
    irc: openstack-meeting
    frequency: biweekly-even
  - time: '2200'
    day: Wednesday
    irc: openstack-meeting
    frequency: biweekly-odd
chair: Jane Developer
description: >
    Weekly meeting for Subteam project.
agenda: |
  * Top bugs this week
"""

ALTERNATING_CHAIRS_MEETING = """
project: OpenStack Subteam Meeting
schedule:
  - time: '1200'
    day: Wednesday
    irc: openstack-meeting
    frequency: biweekly-even
    chair: Jane Developer
  - time: '2200'
    day: Wednesday
    irc: openstack-meeting
    frequency: biweekly-odd
    chair: John Developer
description: >
    Weekly meeting for Subteam project.
agenda: |
  * Top bugs this week
"""

BIWEEKLY_EVEN_MEETING = """
project: OpenStack Subteam 12 Meeting
schedule:
  - time: '2200'
    day: Wednesday
    irc: openstack-meeting
    frequency: biweekly-even
chair: Jane Developer
description: >
    Weekly meeting for Subteam project.
agenda: |
  * Top bugs this week
"""

BIWEEKLY_ODD_MEETING = """
project: OpenStack Subteam 12 Meeting
schedule:
  - time: '2200'
    day: Wednesday
    irc: openstack-meeting
    frequency: biweekly-odd
chair: Jane Developer
description: >
    Weekly meeting for Subteam project.
agenda: |
  * Top bugs this week
"""

MEETING_SUNDAY_LATE = """
project: OpenStack Subteam 8 Meeting
schedule:
  - time: '2330'
    day: Sunday
    irc: openstack-meeting
    frequency: weekly
chair: Shannon Stacker
description: >
    Weekly late meeting for Subteam 8 project.
"""

MEETING_MONDAY_EARLY = """
project: OpenStack Subteam Meeting
schedule:
  - time: '0000'
    day: Monday
    irc: openstack-meeting
    frequency: weekly
chair: Joe Developer
description: >
    Weekly long meeting for Subteam project.
"""

MEETING_MONDAY_LATE = """
project: OpenStack Subteam 8 Meeting
schedule:
  - time: '2330'
    day: Monday
    irc: openstack-meeting
    frequency: weekly
chair: Shannon Stacker
description: >
    Weekly late meeting for Subteam 8 project.
"""

MEETING_TUESDAY_EARLY = """
project: OpenStack Subteam Meeting
schedule:
  - time: '0000'
    day: Tuesday
    irc: openstack-meeting
    frequency: weekly
chair: Joe Developer
description: >
    Weekly long meeting for Subteam project.
"""

MEETING_WITH_DURATION = """
project: OpenStack Subteam 8 Meeting
schedule:
  - time: '1200'
    duration: 30
    day: Wednesday
    irc: openstack-meeting
    frequency: weekly
chair: Shannon Stacker
description: >
    Weekly short meeting for Subteam project.
agenda: |
  * Debate whether this should be a longer meeting
"""

MEETING_WITH_START_DATE = """
project: OpenStack Subteam 8 Meeting
schedule:
  - time: '1200'
    duration: 30
    day: Thursday
    start_date: 20150801
    irc: openstack-meeting
    frequency: weekly
chair: Shannon Stacker
description: >
    Weekly short meeting for Subteam project.
agenda: |
  * Debate whether this should be a longer meeting
"""

MEETING_WITH_SKIP_DATES = """
project: OpenStack Subteam 8 Meeting
schedule:
  - time: '1200'
    day: Monday
    start_date: 20150801
    irc: openstack-meeting
    frequency: weekly
    skip_dates:
        - skip_date: 20150810
          reason: Chair on vacation
chair: Shannon Stacker
description: >
    Weekly short meeting for Subteam project.
"""

MEETING_WITH_SKIP_DATES_BAD_DATE = """
project: OpenStack Subteam 8 Meeting
schedule:
  - time: '1200'
    day: Monday
    start_date: 20150801
    irc: openstack-meeting
    frequency: weekly
    skip_dates:
        - skip_date: 2015080
          reason: Chair on vacation
chair: Shannon Stacker
description: >
    Weekly short meeting for Subteam project.
"""

# typo in skip_date
MEETING_WITH_MISSING_SKIP_DATE = """
project: OpenStack Subteam 8 Meeting
schedule:
  - time: '1200'
    day: Monday
    start_date: 20150801
    irc: openstack-meeting
    frequency: weekly
    skip_dates:
        - skiip_date: 20150806
          reason: Chair on vacation
chair: Shannon Stacker
description: >
    Weekly short meeting for Subteam project.
"""

# typo in reason
MEETING_WITH_MISSING_REASON = """
project: OpenStack Subteam 8 Meeting
schedule:
  - time: '1200'
    day: Monday
    start_date: 20150801
    irc: openstack-meeting
    frequency: weekly
    skip_dates:
        - skip_date: 20150806
          reaso: Chair on vacation
chair: Shannon Stacker
description: >
    Weekly short meeting for Subteam project.
"""

# adhoc meeting
ADHOC_MEETING = """
project: OpenStack Random Meeting
schedule:
  - time: '1200'
    day: Monday
    start_date: 20150801
    irc: openstack-meeting
    frequency: adhoc
chair: Shannon Stacker
description: >
    Adhoc random meeting for Subteam project.
"""

QUADWEEKLY_MEETING_ALTERNATING = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Wednesday
    irc: openstack-meeting
    frequency: quadweekly
  - time: '600'
    duration: 45
    start_date: 20150801
    day: Thursday
    irc: openstack-meeting
    frequency: quadweekly-alternate
chair: John Doe
description:  >
  Example alternating quadweekly meeting
"""

QUADWEEKLY_MEETING = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Wednesday
    irc: openstack-meeting
    frequency: quadweekly
chair: John Doe
description:  >
  Example Quadweekly meeting
"""

QUADWEEKLY_MEETING_WEEK_1 = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Wednesday
    irc: openstack-meeting
    frequency: quadweekly-week-1
chair: John Doe
description:  >
  Example Quadweekly meeting on week 1
"""

QUADWEEKLY_MEETING_WEEK_2 = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Wednesday
    irc: openstack-meeting
    frequency: quadweekly-week-2
chair: John Doe
description:  >
  Example Quadweekly meeting on week 2
"""

QUADWEEKLY_MEETING_WEEK_3 = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Wednesday
    irc: openstack-meeting
    frequency: quadweekly-week-3
chair: John Doe
description:  >
  Example Quadweekly meeting on week 3
"""

QUADWEEKLY_MEETING_ALTERNATE = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Wednesday
    irc: openstack-meeting
    frequency: quadweekly-alternate
chair: John Doe
description:  >
  Example Quadweekly Alternate meeting
"""

FIRST_MONDAY_MEETING = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Monday
    irc: openstack-meeting
    frequency: first-monday
chair: John Doe
description:  >
  Example Monthly meeting
"""

FIRST_TUESDAY_MEETING = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Tuesday
    irc: openstack-meeting
    frequency: first-tuesday
chair: John Doe
description:  >
  Example Monthly meeting
"""

WEEKLY_MEETING_2200 = """
project: OpenStack Subteam Meeting
schedule:
  - time: '2200'
    day: Wednesday
    irc: openstack-meeting
    frequency: weekly
chair: Joe Developer
description: >
    Weekly meeting for Subteam project.
agenda: |
  * Top bugs this week
"""

FIRST_WEDNESDAY_MEETING = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Wednesday
    irc: openstack-meeting
    frequency: first-wednesday
chair: John Doe
description:  >
  Example Monthly meeting
"""

FIRST_THURSDAY_MEETING = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Thursday
    irc: openstack-meeting
    frequency: first-thursday
chair: John Doe
description:  >
  Example Monthly meeting
"""

FIRST_FRIDAY_MEETING = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Friday
    irc: openstack-meeting
    frequency: first-friday
chair: John Doe
description:  >
  Example Monthly meeting
"""

SECOND_WEDNESDAY_MEETING = """
project: CentOS Cloud SIG
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Wednesday
    irc: centos-meeting
    frequency: second-wednesday
chair: Jane Deer
description:  >
  Example Monthly meeting
"""

SECOND_THURSDAY_MEETING = """
project: CentOS Storage SIG
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Thursday
    irc: centos-meeting
    frequency: second-thursday
chair: Jane Deer
description:  >
  Example Monthly meeting
"""

SECOND_FRIDAY_MEETING = """
project: CentOS Hyperscale SIG
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Friday
    irc: centos-meeting
    frequency: second-friday
chair: Jane Deer
description:  >
  Example Monthly meeting
"""

THIRD_MONDAY_MEETING = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Monday
    irc: openstack-meeting
    frequency: third-monday
chair: John Doe
description:  >
  Example Monthly meeting
"""

THIRD_TUESDAY_MEETING = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Tuesday
    irc: openstack-meeting
    frequency: third-tuesday
chair: John Doe
description:  >
  Example Monthly meeting
"""

THIRD_WEDNESDAY_MEETING = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Wednesday
    irc: openstack-meeting
    frequency: third-wednesday
chair: John Doe
description:  >
  Example Monthly meeting
"""

THIRD_THURSDAY_MEETING = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Thursday
    irc: openstack-meeting
    frequency: third-thursday
chair: John Doe
description:  >
  Example Monthly meeting
"""

THIRD_FRIDAY_MEETING = """
project: OpenStack Random Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Friday
    irc: openstack-meeting
    frequency: third-friday
chair: John Doe
description:  >
  Example Monthly meeting
"""

FOURTH_MONDAY_MEETING = """
project: CentOS SIG Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Monday
    irc: centos-meeting
    frequency: fourth-monday
chair: Jane Deer
description:  >
  Example Monthly meeting
"""

FOURTH_TUESDAY_MEETING = """
project: CentOS SIG Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Tuesday
    irc: centos-meeting
    frequency: fourth-tuesday
chair: Jane Deer
description:  >
  Example Monthly meeting
"""

FOURTH_WEDNESDAY_MEETING = """
project: CentOS SIG Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Wednesday
    irc: centos-meeting
    frequency: fourth-wednesday
chair: Jane Deer
description:  >
  Example Monthly meeting
"""

FOURTH_THURSDAY_MEETING = """
project: CentOS SIG Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Thursday
    irc: centos-meeting
    frequency: fourth-thursday
chair: Jane Deer
description:  >
  Example Monthly meeting
"""

FOURTH_FRIDAY_MEETING = """
project: CentOS SIG Meeting
agenda_url: http://agenda.com/
project_url: http://project.com
schedule:
  - time: '2200'
    day: Friday
    irc: centos-meeting
    frequency: fourth-friday
chair: Jane Deer
description:  >
  Example Monthly meeting
"""
